#include "../Software.h"
#include "../../../Misc.h"

#include <X11/X.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

static Window window;
static XImage* framebufferImage;
static unsigned char* framebufferPixels;
static int framebufferPitch;
static char* xfriendlyFB;
static GC gc;
static int screenWidth;
static int screenHeight;
static bool useShmP = false;
static XShmSegmentInfo seginfo;

extern Display* xDisplay;
extern Visual* xVisual;

static bool CheckShm() {
    bool exists = XShmQueryExtension(xDisplay);
    return exists;
}

bool checkImageOK(XImage* image) {
    bool ok = true;
    ok &= (image->width == screenWidth);
    ok &= (image->bytes_per_line == screenWidth * 4);
    ok &= (image->depth == 24 || image->depth == 32);
    return ok;
}

void createFramebuffer() {
    gc = DefaultGC(xDisplay, DefaultScreen(xDisplay));

    int depth = DefaultDepth(xDisplay, DefaultScreen(xDisplay));

    // Setting up the framebuffer
    xfriendlyFB = NULL;

    if(useShmP) {
        framebufferImage = XShmCreateImage(xDisplay, xVisual,
                                           depth, ZPixmap, NULL,
                                           &seginfo, screenWidth, screenHeight);
        seginfo.shmid = shmget(IPC_PRIVATE,
                               framebufferImage->bytes_per_line * framebufferImage->height,
                               IPC_CREAT | 0777);
        seginfo.shmaddr = xfriendlyFB = framebufferImage->data = (char*)shmat(seginfo.shmid, NULL, 0);
        XShmAttach(xDisplay, &seginfo);
    } else {
        xfriendlyFB = (char*)malloc(screenWidth * screenHeight * 4);
        framebufferImage = XCreateImage(xDisplay, xVisual,
                                        depth, ZPixmap, 0,
                                        (char*)xfriendlyFB, screenWidth, screenHeight,
                                        32, 0);
    }
}

bool WindowBackend_Software_CreateWindow(const char *window_title, size_t screen_width, size_t screen_height, bool fullscreen) {
    screenWidth = screen_width;
    screenHeight = screen_height;

    int screen = DefaultScreen(xDisplay);
    Window rootWindow = RootWindow(xDisplay, screen);

    useShmP = CheckShm();
    if(useShmP) {
        Backend_PrintInfo("Using X shared memory extension for higher image transfer speeds.\n");
    } else {
        Backend_PrintInfo("Unable to use X shared memory extension. Expect lower performance.\n");
        Backend_PrintInfo("On Sun, you might need to do the following:\n\nsetenv DISPLAY :0\nsetenv XSUNTRANSPORT shmem\nsetenv XSUNSMESIZE 512\n");
    }

    window = XCreateWindow(xDisplay, rootWindow, 0, 0, screen_width, screen_height, 0, 24, InputOutput, xVisual, 0, NULL);
    XSelectInput(xDisplay, window, ExposureMask | KeyPressMask | KeyReleaseMask);
    XMapRaised(xDisplay, window);
    XStoreName(xDisplay, window, window_title);

    // Make the close button work. This is also required for the window to show up.
    // weird
    Atom WM_DELETE_WINDOW = XInternAtom(xDisplay, "WM_DELETE_WINDOW", false);
    XSetWMProtocols(xDisplay, window, &WM_DELETE_WINDOW, 1);

    createFramebuffer();
    if(checkImageOK(framebufferImage) == false) {
        Backend_PrintError("Something's off with the framebuffer.\n");
        return false;
    }

    if(framebufferImage == 0) {
        return false;
    }

    framebufferPitch = screen_width * 4;
    return true;
}

unsigned char* WindowBackend_Software_GetFramebuffer(size_t *pitch)
{
	*pitch = framebufferPitch;

	return (unsigned char*)xfriendlyFB;
}

void WindowBackend_Software_DestroyWindow(void) {
    if(useShmP) {
        XShmDetach(xDisplay, &seginfo);
        XDestroyImage(framebufferImage);
        shmdt(seginfo.shmaddr);
        shmctl(seginfo.shmid, IPC_RMID, 0);
    }
    XDestroyWindow(xDisplay, window);
}

void WindowBackend_Software_HandleWindowResize(size_t width, size_t height) {
	(void)width;
	(void)height;

    Backend_PrintError("Bro I can't resize, I'm too lazy.");
}

void WindowBackend_Software_Display(void) {
    // Convert Cave Story's framebuffer format to the Sun's
    // This is the exact reverse of how it is on modern
    // computers. Yup, it's endianess.
    // int srci = 0;
    // for(int i = 0; i < screenWidth * screenHeight * 4; i += 4) {
    //     int src;
    //     xfriendlyFB[i + 3] = 0; // UNUSED
    //     xfriendlyFB[i + 2] = framebufferPixels[srci + 0]; // BLUE
    //     xfriendlyFB[i + 1] = framebufferPixels[srci + 1]; // GREEN
    //     xfriendlyFB[i + 0] = framebufferPixels[srci + 2]; // RED
    //     srci += 4;
    // }
    //memcpy(xfriendlyFB, framebufferPixels, screenWidth * screenHeight * 4);
    if(useShmP) {
        XShmPutImage(xDisplay, window, gc, framebufferImage, 0, 0, 0, 0, screenWidth, screenHeight, 0);
    } else {
        XPutImage(xDisplay, window, gc, framebufferImage, 0, 0, 0, 0, screenWidth, screenHeight);
    }
}
